"""Tests functions in `plots`."""

import unittest
from contextlib import contextmanager

import matplotlib.pyplot as plt
import numpy as np

from aind_mri_utils import plots as mrplt


@contextmanager
def managed_subplot(f, *args, **kwargs):
    ax = f.add_subplot(*args, **kwargs)
    try:
        yield ax
    finally:
        f.clf()


class PlotsTest(unittest.TestCase):
    """Tests functions in `plots`."""

    f1 = plt.figure(1)

    vertices = np.array(
        [
            (-6.859, 0.7264, 2.84),
            (-6.859, 1.5264, 2.84),
            (-8.059, 1.5264, 2.84),
            (-8.049885, 1.5264, 2.944189),
            (-8.022816, 1.5264, 3.045212),
        ]
    )
    faces = np.array(
        [
            [0, 2, 3],
            [0, 3, 4],
            [2, 3, 4],
            [1, 2, 4],
            [1, 3, 4],
        ]
    )

    expected_edges = np.array(
        [
            [2, 0],
            [2, 1],
            [3, 0],
            [3, 1],
            [3, 2],
            [4, 0],
            [4, 1],
            [4, 2],
            [4, 3],
        ],
        dtype="int32",
    )

    def test_make_3d_ax_look_normal(self) -> None:
        """Tests make_3d_ax_look_normal"""
        with managed_subplot(self.f1, projection="3d") as ax3d:
            mrplt.make_3d_ax_look_normal(ax3d)
            box_aspect = ax3d.get_box_aspect()
            self.assertTrue(
                np.array_equal(
                    box_aspect / box_aspect[0], np.ones(3, dtype="float64")
                )
            )

    def test_set_axes_equal(self) -> None:
        """Tests set_axes_equal"""
        with managed_subplot(self.f1, projection="3d") as ax3d:
            mrplt.set_axes_equal(ax3d)
            limits = np.array(
                [
                    ax3d.get_xlim3d(),
                    ax3d.get_ylim3d(),
                    ax3d.get_zlim3d(),
                ]
            )
            limits_diff = np.diff(limits)
            self.assertTrue(np.all(limits_diff == limits_diff[0]))

    def test_plot_tri_mesh(self) -> None:
        """Tests plot_tri_mesh"""
        with managed_subplot(self.f1, projection="3d") as ax3d:
            handles, tri = mrplt.plot_tri_mesh(ax3d, self.vertices, self.faces)
            self.assertTrue(np.array_equal(tri.edges, self.expected_edges))

    def test_plot_point_cloud_3d(self) -> None:
        with managed_subplot(self.f1, projection="3d") as ax3d:
            outs = mrplt.plot_point_cloud_3d(ax3d, self.faces)
            self.assertTrue(outs.get_offsets().size == 10)

    def test_plot_vector(self) -> None:
        with managed_subplot(self.f1, projection="3d") as ax3d:
            outs = mrplt.plot_vector(ax3d, self.faces[0, :])
            self.assertTrue(len(outs) == 1)

    def create_single_colormap(self) -> None:
        """Tests create_single_color_map"""
        # Test colormap creation
        A = mrplt.create_single_colormap("magenta", start_color="white")
        self.assertTrue(np.array_equal(A(0), np.array([1, 1, 1, 1])))
        self.assertTrue(np.array_equal(A(1), np.array([1, 0, 1, 1])))
        # Test colormap creation with reverse
        A = mrplt.create_single_colormap(
            "magenta", start_color="white", is_reverse=True
        )
        self.assertTrue(np.array_equal(A(0), np.array([1, 0, 1, 1])))
        self.assertTrue(np.array_equal(A(1), np.array([1, 1, 1, 1])))
        # Test output length
        A = mrplt.create_single_colormap(
            "magenta", start_color="white", N=4192
        )
        self.assertTrue(len(A) == 4192)
        # Test transparency
        A = mrplt.create_single_colormap(
            "magenta", start_color="white", N=4192, is_transparent=True
        )
        self.assertTrue(np.array_equal(A(0), np.array([1, 1, 1, 0])))
        # Tst saturation
        A = mrplt.create_single_colormap(
            "magenta", start_color="white", saturation=4
        )
        self.assertTrue(np.array_equal(A(254), np.array([1, 0, 1, 1])))


if __name__ == "__main__":
    unittest.main()



"""
Code for handling manual keypoints transforms using vtk
"""

import numpy as np
import vtk


def define_transform(source_landmarks, target_landmarks):
    """
    Defines a non-linear warp between a set of source and target landmarks

    Parameters
    ==========
    source_landmarks - np.ndarray (N x 3)
    target_landmarks - np.ndarray (N x 3)

    Returns
    =======
    transform - vtkThinPlateSplineTransform

    """

    transform = vtk.vtkThinPlateSplineTransform()
    source_points = vtk.vtkPoints()
    target_points = vtk.vtkPoints()

    for i in range(source_landmarks.shape[0]):
        source_points.InsertNextPoint(source_landmarks[i, :])

    for i in range(target_landmarks.shape[0]):
        target_points.InsertNextPoint(target_landmarks[i, :])

    transform.SetBasisToR()  # for 3D transform
    transform.SetSourceLandmarks(source_points)
    transform.SetTargetLandmarks(target_points)
    transform.Update()

    return transform


def apply_transform(transform, points):
    """
    Applies a non-linear warp to a set of points

    Parameters
    ==========
    transform - vtkThinPlateSplineTransform
    points - np.ndarray (N x 3)

    Returns
    =======
    warped_points - np.ndarray (N x 3)

    """

    warped_points = np.zeros(points.shape)
    for i in range(points.shape[0]):
        warped_points[i, :] = transform.TransformPoint(points[i, :])

    return warped_points



"""Tests functions in `utils`."""

import unittest

import numpy as np

from aind_mri_utils import utils as ut


class UtilsTest(unittest.TestCase):
    """Tests functions in `utls`."""

    cross_product_sets = [
        (np.array([0, 1, 1]), np.array([1, 0, 0]), np.array([0, 1, -1])),
        (np.array([1, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1])),
        (np.array([0, 1, 0]), np.array([1, 0, 0]), np.array([0, 0, -1])),
        (np.array([1, 0, 0]), np.array([1, 0, 0]), np.array([0, 0, 0])),
    ]

    test_match_arr = np.arange(0, 9).reshape(3, 3)
    test_proj_vec = np.array([0.5, 0.2, 0.7])

    signed_angle_sets = [
        (
            np.array([1, 0, 0]),
            np.array([np.pi, np.pi, 0]),
            np.array([0, 0, 1]),
            0.785398,
        ),
        (
            np.array([np.pi, np.pi, 0]),
            np.array([1, 0, 0]),
            np.array([0, 0, 1]),
            -0.785398,
        ),
        (np.array([1, 0, 0]), np.array([1, 0, 0]), np.array([0, 0, 1]), 0.0),
    ]

    def test_skew_symmetric_cross_product_matrix(self) -> None:
        for a, b, c in self.cross_product_sets:
            self.assertTrue(
                np.allclose(ut.skew_symmetric_cross_product_matrix(a) @ b, c)
            )

    def test_norm_vec(self) -> None:
        with self.assertRaises(ValueError):
            ut.norm_vec(np.array([0, 0, 0]))
        self.assertTrue(
            np.allclose(ut.norm_vec(np.array([2, 0, 0])), np.array([1, 0, 0]))
        )

    def test_find_indices_equal_to(self) -> None:
        self.assertTrue(
            np.array_equal(
                ut.find_indices_equal_to(self.test_match_arr, 1),
                np.array([[0, 1]]),
            )
        )
        self.assertEqual(
            len(ut.find_indices_equal_to(self.test_match_arr, 9)), 0
        )
        B = np.copy(self.test_match_arr)
        B[1, :] = 9
        self.assertTrue(
            np.array_equal(
                ut.find_indices_equal_to(B, 9),
                np.array([[1, 0], [1, 1], [1, 2]]),
            )
        )

    def test_vector_rejection(self) -> None:
        for i in range(0, 3):
            ans = np.copy(self.test_proj_vec)
            ans[i] = 0
            basis_vec = np.zeros((3,))
            basis_vec[i] = 1
            self.assertTrue(
                np.allclose(
                    ut.vector_rejection(self.test_proj_vec, basis_vec),
                    ans,
                )
            )

    def test_mask_arr_by_annotation(self) -> None:
        ans = ut.mask_arr_by_annotations(
            self.test_match_arr, self.test_match_arr, [9]
        )
        self.assertTrue(np.allclose(ans, np.zeros((3, 3))))
        ans = ut.mask_arr_by_annotations(
            self.test_match_arr, self.test_match_arr, [1]
        )
        self.assertTrue(np.isclose(self.test_match_arr[0, 1], ans[0, 1]))
        modified_ans = np.copy(ans)
        modified_ans[0, 1] = 0
        self.assertTrue(np.allclose(modified_ans, np.zeros((3, 3))))

    def test_signed_angle(self) -> None:
        for a, b, n, ans in self.signed_angle_sets:
            received = ut.signed_angle_rh(a, b, n)
            self.assertTrue(np.isclose(received, ans))
            self.assertTrue(np.isclose(ans, ut.signed_angle_lh(b, a, n)))

    def test_get_first_pca_axis(self) -> None:
        pts = np.array([[1, 2], [3, 4], [5, 6]])
        expected_first_pc = np.array([0.70710678, 0.70710678])
        self.assertTrue(
            np.allclose(
                ut.get_first_pca_axis(pts),
                expected_first_pc,
            )
        )

    def test_unsigned_angle(self) -> None:
        for a, b, n, ans in self.signed_angle_sets:
            received = ut.unsigned_angle(a, b)
            self.assertTrue(np.isclose(received, abs(ans)))


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np

from aind_mri_utils.file_io import simpleitk as si


class SITKTest(unittest.TestCase):
    """Tests functions in `file_io.simpleitk`."""

    def test_save_sitk_transform(self) -> None:
        """
        Tests that the `save_sitk_transform` function works as intended.
        """
        # Test that a 4x3 transform is created correctly
        test_save_path = "testsave.h5"
        R = np.eye(3)
        translation = np.zeros(3)
        si.save_sitk_transform(test_save_path, R, translation)
        load_trans, load_translation, load_center = si.load_sitk_transform(
            test_save_path
        )
        self.assertTrue(np.allclose(R, load_trans))

        # Test that the above can work from 6x1
        si.save_sitk_transform(test_save_path, np.zeros(6))
        load_trans, _, _ = si.load_sitk_transform(test_save_path)
        self.assertTrue(np.allclose(R, load_trans))

        # Test more complicated transform
        R[0, -1] = 1
        si.save_sitk_transform(test_save_path, R)
        load_trans, _, _ = si.load_sitk_transform(test_save_path)
        self.assertTrue(np.allclose(R, load_trans))

        # Test inversion functionality when saving
        si.save_sitk_transform(
            test_save_path, R, transpose_matrix=True, legacy=True
        )
        load_trans, _, _ = si.load_sitk_transform(test_save_path)
        load_trans[:3, :3] = load_trans[:3, :3].T
        self.assertTrue(np.array_equal(R, load_trans))

        # Kill the file we created- it was just a test
        os.remove("testsave.h5")


if __name__ == "__main__":
    unittest.main()



"""Functions for working with obj files"""

from pathlib import Path

import numpy as np
import pywavefront


def get_vertices_and_faces(scene):
    """
    Collect vertices and faces for a pywavefront object

    Parameters
    ==========
    scene - a Wavefront object made with `collect_materials=True`

    Returns
    =======
    vertices - np.ndarray (floating point N x 3) array of N 3d points
    faces - list[np.ndarray], list of (int M x 3) array of M triangles, where
            each element corresponds to the index of a vertex in `vertices`.
            Each list element corresponds to a different mesh.
    """
    if isinstance(scene, str) or isinstance(scene, Path):
        scene = load_obj_wavefront(scene)
    vertices = np.array(scene.vertices)
    faces = []
    for mesh in scene.mesh_list:
        faces.append(np.array(mesh.faces))
    return vertices, faces


def load_obj_wavefront(filename):
    """
    Wrapper for loading a pywavefront scene

    Parameters
    ==========
    filename - name of file to load

    Returns
    =======
    scene - scene with default params so I don't need to remember to type them
    """
    scene = pywavefront.Wavefront(
        filename, strict=False, create_materials=True, collect_faces=True
    )
    return scene



import json
import numpy as np



def read_neuroglancer_probes_and_annotations(filename,
                                 probe_layers = None,
                                 annotation_layers = None):
    
    """
    Reads a Neuroglancer JSON file and extracts probe locations and annotation data.
    
    This function parses a Neuroglancer JSON file that contains both probe and annotation layers,
    extracts relevant information, and returns the probe coordinates, annotation coordinates,
    voxel spacing, and dimension order. The probe and annotation layers are either automatically 
    determined from the file or can be specified by the user.
    
    Function written by Yoni, Docstring written by ChatGPT.
    
    Parameters:
    -----------
    filename : str
        Path to the Neuroglancer JSON file to read.
    
    probe_layers : list, optional
        List of specific probe layers to extract. If None, it will automatically extract 
        all layers whose names are numeric strings (default is None).
        Use flag -1 to skip.
    
    annotation_layers : list, optional
        List of specific annotation layers to extract. If None, it will automatically extract 
        all layers not used as probe layers (default is None).
        Use flag -1 to skip.
    
    Returns:
    --------
    probes : dict
        A dictionary where the keys are the probe layer names (strings) and the values are 
        NumPy arrays of the probe coordinates, with the shape (N, 3) where N is the number of 
        probe points in that layer.
    
    annotations : dict
        A dictionary where the keys are annotation descriptions (strings) and the values are 
        NumPy arrays of the annotation coordinates, with the shape (3,).
    
    spacing : numpy.ndarray
        A NumPy array representing the voxel spacing in the x, y, z dimensions.
    
    dimension_order : numpy.ndarray
        A NumPy array containing the dimension order in x, y, z format for reordering.
    
    Notes:
    ------
    - The function assumes that the dimensions in the Neuroglancer file are ordered in a certain 
      way and reorders the data to match the 'xyz' format for consistency.
    - Probes are expected to be defined in layers where the layer name is numeric.
    - If no probe or annotation layers are provided, the function will attempt to autodetect them.
    """

    # Read the json file into memory
    with open(filename) as f:
        data = json.load(f)
        
    # Get the dimension order/spacing for file.
    dimension_order = list(data['dimensions'].keys())[:3]
    spacing = np.zeros(len(dimension_order))
    for ii,key in enumerate(dimension_order):
        spacing[ii] = data['dimensions'][key][0]
    # For reordering data into xyz
    dim_order = list(np.argsort(dimension_order))

        
    layers = data['layers']
    probes = {}

    if probe_layers==-1:
        probes = False
    elif probe_layers == None:
        for ii,this_layer in enumerate(layers):
            if this_layer['type']!='annotation':
                continue
            if this_layer['name'].isdigit():
                probes[this_layer['name']] = []
                for jj,this_point in enumerate(this_layer['annotations']):
                    probes[this_layer['name']].append(this_point['point'][:-1])
                probes[this_layer['name']]  = np.array(probes[this_layer['name']])*spacing
                probes[this_layer['name']] = probes[this_layer['name']][:,dim_order]
    else:
        for ii,this_layer in enumerate(probe_layers):
            probes[this_layer['name']] = []
            for jj,this_point in enumerate(this_layer['annotations']):
                probes[this_layer['name']].append(this_point['point'][:-1])
            probes[this_layer['name']]  = np.array(probes[this_layer['name']])*spacing
            probes[this_layer['name']] = probes[this_layer['name']][:,dim_order]
    
    
    annotations = {}
    if annotation_layers==-1:
        annotations = False
    elif annotation_layers == None:
        if probe_layers==None:
            probe_layers = list(probes.keys())
        for ii,this_layer in enumerate(layers):
            if (this_layer['type']!='annotation') or (this_layer['name'] in probe_layers):
                continue
            else:
                print(this_layer['name'])

                for jj,this_point in enumerate(this_layer['annotations']):
                    annotations[this_point['description'][:-1]] = np.array(this_point['point'][:-1])*spacing
                    annotations[this_point['description'][:-1]] = annotations[this_point['description'][:-1]][dim_order]
    else:
        for ii,this_layer in enumerate(annotation_layers):
            for jj,this_point in enumerate(this_layer['annotations']):
                annotations[this_point['description'][:-1]] = np.array(this_point['point'][:-1])*spacing
                annotations[this_point['description'][:-1]] = annotations[this_point['description'][:-1]][dim_order]

    return probes,annotations,spacing[dim_order],np.array(dimension_order)[dim_order]


# -*- coding: utf-8 -*-
"""
Functions for Loading and manipulating meshes durring insertion planning.
"""

import trimesh


def as_mesh(scene_or_mesh):
    """
    Convert a possible scene to a mesh.

    If conversion occurs, the returned mesh has only vertex and face data.

    see https://github.com/mikedh/trimesh/issues/507
    """
    if isinstance(scene_or_mesh, trimesh.Scene):
        if len(scene_or_mesh.geometry) == 0:
            mesh = None  # empty scene
        else:
            # we lose texture information here
            mesh = trimesh.util.concatenate(
                tuple(
                    trimesh.Trimesh(vertices=g.vertices, faces=g.faces)
                    for g in scene_or_mesh.geometry.values()
                )
            )
    else:
        assert isinstance(scene_or_mesh, trimesh.Trimesh)
        mesh = scene_or_mesh
    return mesh


def load_newscale_trimesh(
    filename,
    move_down=0,
):
    """
    Load a newscale model mesh

    """
    mesh = trimesh.load_mesh(filename)
    mesh = as_mesh(mesh)
    mesh.vertices = mesh.vertices[:, [0, 2, 1]]  # Should be made into params.
    mesh.vertices[:, 2] = mesh.vertices[:, 2] - move_down
    # Repair broken stuff from bad blender-ing
    trimesh.repair.broken_faces(mesh)
    trimesh.repair.fix_normals(mesh)
    trimesh.repair.fix_inversion(mesh)
    trimesh.repair.fix_winding(mesh)
    return mesh


def apply_transform_to_trimesh(mesh, T):
    """
    Apply a transform to a trimesh Mesh object
    """
    mesh.vertices = trimesh.transform_points(mesh.vertices, T)
    return mesh


def create_uv_spheres(positions, radius=0.25, color=[255, 0, 255, 255]):
    """
    Create UV spheres at specified positions with a given radius and color.

    Parameters
    ----------
    positions : ndarray
        Array of positions where spheres should be created.
    radius : float, optional
        Radius of the spheres, by default 0.25.
    color : list, optional
        RGBA color of the spheres, by default [255, 0, 255, 255].

    Returns
    -------
    list
        List of trimesh objects representing the spheres.
    """
    meshes = [
        trimesh.creation.uv_sphere(radius=radius)
        for _ in range(len(positions))
    ]
    for i, mesh in enumerate(meshes):
        mesh.apply_translation(positions[i, :])
        mesh.visual.vertex_colors = color
    return meshes
